@@ -3,9 +3,8 @@
 Alignment module.
 """
 from collections import namedtuple
-import math
-from ._align import Aligner, NoIndelAligner, compare_prefixes, locate
-from .util import reverse_complement
+from ._align import Aligner, compare_prefixes, locate, MultiAligner
+from .util import reverse_complement, RandomMatchProbability
 
 # flags for global alignment
 
@@ -159,51 +158,34 @@ MatchInfo = namedtuple("MatchInfo", (
 #    http://www.nature.com/articles/srep01746
 #    http://www.isical.ac.in/~bioinfo_miu/FOGSAA.7z
 
-class FactorialCache(object):
-    def __init__(self, init_size=150):
-        self.factorials = [1] * init_size
-        self.max_n = 1
-        self.cur_array_size = init_size
-
-    def factorial(self, n):
-        if n > self.max_n:
-            self._fill_upto(n)
-        return self.factorials[n]
-
-    def _fill_upto(self, n):
-        if n >= self.cur_array_size:
-            extension_size = n - self.cur_array_size + 1
-            self.factorials += [1] * extension_size
-        i = self.max_n
-        next_i = i + 1
-        while i < n:
-            self.factorials[next_i] = next_i * self.factorials[i]
-            i = next_i
-            next_i += 1
-        self.max_n = i
-
 class InsertAligner(object):
     """
     Implementation of an insert matching algorithm.
     This only works with paired-end reads with 3' adapters.
     """
-    def __init__(self, adapter1, adapter2, max_error_prob=1E-6,
+    def __init__(self, adapter1, adapter2, match_probability=RandomMatchProbability(),
+                 insert_max_rmp=1E-6, adapter_max_rmp=0.001,
                  min_insert_overlap=1, max_insert_mismatch_frac=0.2,
-                 min_adapter_overlap=1, min_adapter_match_frac=0.8,
-                 adapter_check_cutoff=9):
+                 min_adapter_overlap=1, max_adapter_mismatch_frac=0.2,
+                 adapter_check_cutoff=9, base_probs=dict(p1=0.25, p2=0.75)):
         self.adapter1 = adapter1
         self.adapter1_len = len(adapter1)
         self.adapter2 = adapter2
         self.adapter2_len = len(adapter2)
-        self.max_error_prob = max_error_prob
+        self.match_probability = match_probability
+        self.insert_max_rmp = insert_max_rmp
+        self.adapter_max_rmp = adapter_max_rmp
         self.min_insert_overlap = min_insert_overlap
         self.max_insert_mismatch_frac = float(max_insert_mismatch_frac)
         self.min_adapter_overlap = min_adapter_overlap
-        self.min_adapter_match_frac = float(min_adapter_match_frac)
-        self.max_adapter_mismatch_frac = 1.0 - self.min_adapter_match_frac
+        self.max_adapter_mismatch_frac = float(max_adapter_mismatch_frac)
         self.adapter_check_cutoff = adapter_check_cutoff
-        self.factorial_cache = FactorialCache()
-
+        self.base_probs = base_probs
+        self.aligner = MultiAligner(
+            max_insert_mismatch_frac,
+            START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2,
+            min_insert_overlap)
+    
     def match_insert(self, seq1, seq2):
         """Use cutadapt aligner for insert and adapter matching"""
         l1 = len(seq1)
@@ -215,69 +197,96 @@ class InsertAligner(object):
             seq2 = seq1[:l1]
 
         seq2_rc = reverse_complement(seq2)
-        result = [None, 0, None, None]
         
-        aligner = Aligner(
-            seq2_rc,
-            self.max_insert_mismatch_frac,
-            START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2,
-            False, False)
-        aligner.min_overlap = self.min_insert_overlap
-        aligner.indel_cost = 100000
-        
-        # In practice, this is only faster at relatively high error rates. I want to
-        # write more unit tests, and possible try a different algorithm instead, before
-        # enable this.
-        #aligner = NoIndelAligner(
-        #    seq2_rc,
-        #    self.max_insert_mismatch_frac,
-        #    START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2)
-        #aligner.min_overlap = self.min_insert_overlap
-        
-        insert_match = aligner.locate(seq1)
-        
-        if insert_match:
-            offset = min(insert_match[0], seq_len - insert_match[3])
-            insert_match_size = seq_len - offset
-            result[0] = insert_match
-            result[1] = insert_match_size
-
-            if (offset < self.min_adapter_overlap or
-                    self.match_probability(insert_match[4], insert_match_size) > self.max_error_prob):
-                return result
-
+        def _match(insert_match, offset, insert_match_size, prob):
+            if offset < self.min_adapter_overlap:
+                # The reads are mostly overlapping, to the point where
+                # there's not enough overhang to do a confident adapter
+                # match. We return just the insert match to signal that
+                # error correction can be done even though no adapter
+                # trimming is required.
+                return (insert_match, None, None)
+            
+            # TODO: this is very sensitive to the exact correct choice of adapter.
+            # For example, if you specifiy GATCGGAA... and the correct adapter is
+            # AGATCGGAA..., the prefixes will not match exactly and the alignment
+            # will fail. We need to use a comparison that is a bit more forgiving.
+            
             a1_match = compare_prefixes(seq1[insert_match_size:], self.adapter1)
             a2_match = compare_prefixes(seq2[insert_match_size:], self.adapter2)
             adapter_len = min(offset, self.adapter1_len, self.adapter2_len)
-            min_adapter_matches = math.ceil(adapter_len * self.min_adapter_match_frac)
-            if a1_match[4] < min_adapter_matches and a2_match[4] < min_adapter_matches:
-                return result
+            max_adapter_mismatches = round(adapter_len * self.max_adapter_mismatch_frac)
+            if a1_match[5] > max_adapter_mismatches and a2_match[5] > max_adapter_mismatches:
+                return None
+            
             a1_prob = self.match_probability(a1_match[4], adapter_len)
             a2_prob = self.match_probability(a2_match[4], adapter_len)
-            if adapter_len > self.adapter_check_cutoff and (a1_prob * a2_prob) > self.max_error_prob:
-                return result
+            if (adapter_len > self.adapter_check_cutoff) and ((a1_prob * a2_prob) > self.adapter_max_rmp):
+                return None
 
             adapter_len1 = min(self.adapter1_len, l1 - insert_match_size)
             adapter_len2 = min(self.adapter2_len, l2 - insert_match_size)
             best_adapter_matches, best_adapter_mismatches = (a1_match if a1_prob < a2_prob else a2_match)[4:6]
-            result[2] = Match(0, adapter_len1, insert_match_size, l1, best_adapter_matches, best_adapter_mismatches)
-            result[3] = Match(0, adapter_len2, insert_match_size, l2, best_adapter_matches, best_adapter_mismatches)
-
-        return result
-    
-    def match_probability(self, matches, size):
-        nfac = self.factorial_cache.factorial(size)
-        p = 0.0
-        i = matches
-
-        while i <= size:
-            p += (
-                (0.75 ** (size - i)) *
-                (0.25 ** i) *
-                nfac /
-                self.factorial_cache.factorial(i) /
-                self.factorial_cache.factorial(size - i)
+            
+            return (
+                insert_match,
+                Match(0, adapter_len1, insert_match_size, l1, best_adapter_matches, best_adapter_mismatches),
+                Match(0, adapter_len2, insert_match_size, l2, best_adapter_matches, best_adapter_mismatches)
             )
-            i += 1
-
-        return p
+        
+        # # This is the old way of doing things, where we use the built-in
+        # # Aligner to do a single match.
+        # aligner = Aligner(
+        #     seq2_rc,
+        #     self.max_insert_mismatch_frac,
+        #     START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2,
+        #     False, False)
+        # aligner.min_overlap = self.min_insert_overlap
+        # aligner.indel_cost = 100000
+        #
+        # insert_match = aligner.locate(seq1)
+        #
+        # if not insert_match:
+        #     return None
+        #
+        # offset = min(insert_match[0], seq_len - insert_match[3])
+        # insert_match_size = seq_len - offset
+        # prob = self.match_probability(insert_match[4], insert_match_size)
+        #
+        # if prob > self.insert_max_rmp:
+        #     return None
+        #
+        # return _match(insert_match, offset, insert_match_size, prob)
+        
+        # Use an aligner that returns all matches that satisfy the
+        # overlap and error rate thresholds. We sort by matches and
+        # then mismatches, and then check each in turn until we find
+        # one with an adapter match (if any).
+        
+        insert_matches = self.aligner.locate(seq2_rc, seq1)
+        
+        if insert_matches:
+            # Filter by random-match probability
+            filtered_matches = []
+            for insert_match in insert_matches:
+                offset = min(insert_match[0], seq_len - insert_match[3])
+                insert_match_size = seq_len - offset
+                prob = self.match_probability(insert_match[4], insert_match_size, **self.base_probs)
+                if prob <= self.insert_max_rmp:
+                    filtered_matches.append((insert_match, offset, insert_match_size, prob))
+            
+            if filtered_matches:
+                if len(filtered_matches) == 1:
+                    return _match(*filtered_matches[0])
+                else:
+                    # Test matches in order of random-match probability.
+                    # TODO: compare against sorting by length (which is how
+                    # SeqPurge essentially does it).
+                    #filtered_matches.sort(key=lambda x: x[2], reverse=True)
+                    filtered_matches.sort(key=lambda x: x[3])
+                    for m in filtered_matches:
+                        match = _match(*m)
+                        if match:
+                            return match
+            
+            return None
