@@ -5,13 +5,12 @@ A modifier must be callable. It is implemented as a function if no parameters
 need to be stored, and as a class with a __call__ method if there are parameters
 (or statistics).
 """
+import copy
+import logging
 import re
 from .qualtrim import quality_trim_index, nextseq_trim_index
 from .align import Aligner, SEMIGLOBAL, InsertAligner
-from .util import complement, reverse_complement
-from collections import defaultdict
-import copy
-import logging
+from .util import complement, reverse_complement, mean
 
 # Base classes
 
@@ -36,7 +35,7 @@ class Trimmer(object):
             return read
     
     def clip(self, read, front=0, back=0):
-        if front or back:
+        if (front or back) and len(read) > 0:
             front_bases, back_bases, new_read = read.clip(front, back)
             self.trimmed_bases += front_bases + back_bases
             return new_read
@@ -94,6 +93,9 @@ class AdapterCutter(object):
 
         Cut found adapters from a single read. Return modified read.
         """
+        if len(read) == 0:
+            return read
+        
         matches = []
 
         # try at most self.times times to remove an adapter
@@ -147,10 +149,13 @@ class InsertAdapterCutter(ReadPairModifier):
     AdapterCutter that uses InsertAligner to first try to identify
     insert overlap before falling back to semi-global adapter alignment.
     """
-    def __init__(self, adapter1, adapter2, action='trim', mismatch_action=None, symmetric=True, **aligner_args):
+    def __init__(self, adapter1, adapter2, action='trim', mismatch_action=None,
+                 symmetric=True, min_insert_overlap=1, **aligner_args):
         self.adapter1 = adapter1
         self.adapter2 = adapter2
-        self.aligner = InsertAligner(adapter1.sequence, adapter2.sequence, **aligner_args)
+        self.aligner = InsertAligner(adapter1.sequence, adapter2.sequence,
+            min_insert_overlap=min_insert_overlap, **aligner_args)
+        self.min_insert_len = min_insert_overlap
         self.action = action
         self.mismatch_action = mismatch_action
         self.symmetric = symmetric
@@ -159,14 +164,16 @@ class InsertAdapterCutter(ReadPairModifier):
         self.corrected_bp = [0, 0]
     
     def __call__(self, read1, read2):
-        insert_match, insert_match_len, adapter_match1, adapter_match2 = self.aligner.match_insert(
-            read1.sequence, read2.sequence)
+        if len(read1) < self.min_insert_len or len(read2) < self.min_insert_len:
+            return (read1, read2)
         
-        if (self.mismatch_action and insert_match and insert_match[5] > 0 and
-                (adapter_match1 or adapter_match2)):
-            self.correct_errors(read1, read2, insert_match)
+        match = self.aligner.match_insert(read1.sequence, read2.sequence)
         
-        if adapter_match1 is None and adapter_match2 is None:
+        if match:
+            insert_match, adapter_match1, adapter_match2 = match
+            if self.mismatch_action and insert_match[5] > 0:
+                self.correct_errors(read1, read2, insert_match)
+        else:
             adapter_match1 = self.adapter1.match_to(read1)
             adapter_match2 = self.adapter2.match_to(read2)
         
@@ -176,42 +183,34 @@ class InsertAdapterCutter(ReadPairModifier):
             if adapter_match2 is None and adapter_match1:
                 adapter_match2 = adapter_match1.copy()
             
-        if adapter_match1 is not None:
-            adapter_match1.adapter = self.adapter1
-            adapter_match1.read = read1
-            adapter_match1.front = False
-        
-        if adapter_match2 is not None:
-            adapter_match2.adapter = self.adapter2
-            adapter_match2.read = read2
-            adapter_match2.front = False
-        
-        return (self.trim(read1, adapter_match1, 0), self.trim(read2, adapter_match2, 1))
+        return (
+            self.trim(read1, self.adapter1, adapter_match1, 0),
+            self.trim(read2, self.adapter2, adapter_match2, 1)
+        )
     
-    def trim(self, read, match, read_idx):
-        # TODO: refactor so that both adapter cutters use the same function
+    def trim(self, read, adapter, match, read_idx):
         if not match:
             read.match = None
             read.match_info = None
             return read
         
-        if match.rstart < len(read):
-            trimmed_read = match.adapter.trimmed(match)
+        match.adapter = adapter
+        match.read = read
+        match.front = False
+    
+        if self.action is None or match.rstart >= len(read):
+            trimmed_read = read
+        
+        else:
+            trimmed_read = adapter.trimmed(match)
             
-            if self.action == 'trim':
-                # read is already trimmed, nothing to do
-                pass
-            elif self.action == 'mask':
+            if self.action == 'mask':
                 # add N from last modification
                 masked_sequence = trimmed_read.sequence
                 masked_sequence += ('N' * len(read) - len(trimmed_read))
                 # set masked sequence as sequence with original quality
                 trimmed_read.sequence = masked_sequence
                 trimmed_read.qualities = read.qualities
-            elif self.action is None:
-                trimmed_read = read
-        else:
-            trimmed_read = read
         
         trimmed_read.match = match
         trimmed_read.match_info = [match.get_info_record()]
@@ -229,20 +228,28 @@ class InsertAdapterCutter(ReadPairModifier):
         if has_quals:
             r1_qual = list(read1.qualities)
             r2_qual = list(read2.qualities)
-        elif self.mismatch_action == 'best':
+        elif self.mismatch_action in ('liberal', 'conservative'):
             raise Exception("Cannot perform quality-based error correction on reads lacking quality information")
         
+        r1_start = insert_match[2]
+        r1_end = insert_match[3]
         r1_changed = 0
+        r2_start = l2 - insert_match[1]
+        r2_end = l2 - insert_match[0]
         r2_changed = 0
+        quals_equal = []
         
-        for i, j in zip(
-                range(insert_match[2], insert_match[3]),
-                range(l2 - insert_match[0] - 1, l2 - insert_match[1] - 1, -1)):
+        for i, j in zip(range(r1_start, r1_end), range(r2_end - 1, r2_start, -1)):
             b1 = r1_seq[i]
             b2 = complement[r2_seq[j]]
             if b1 == b2:
                 continue
-            if b1 == 'N':
+            if self.mismatch_action == 'N':
+                r1_seq[i] = 'N'
+                r2_seq[j] = 'N'
+                r1_changed += 1
+                r2_changed += 1
+            elif b1 == 'N':
                 r1_seq[i] = b2
                 if has_quals:
                     r1_qual[i] = r2_qual[j]
@@ -252,23 +259,38 @@ class InsertAdapterCutter(ReadPairModifier):
                 if has_quals:
                     r2_qual[j] = r1_qual[i]
                 r2_changed += 1
-            elif self.mismatch_action == 'N':
-                r1_seq[i] = 'N'
-                r2_seq[i] = 'N'
-                r1_changed += 1
-                r2_changed += 1
-            elif r1_qual[i] == r2_qual[j]:
-                # default to read1 base
-                r2_seq[j] = complement[b1]
-                r2_changed += 1
-            elif r1_qual[i] > r2_qual[j]:
-                r2_seq[j] = complement[b1]
-                r2_qual[j] = r1_qual[i]
-                r2_changed += 1
-            else:
-                r1_seq[i] = b2
-                r1_qual[i] = r2_qual[j]
-                r1_changed += 1
+            elif has_quals:
+                if r1_qual[i] > r2_qual[j]:
+                    r2_seq[j] = complement[b1]
+                    r2_qual[j] = r1_qual[i]
+                    r2_changed += 1
+                elif r2_qual[j] > r1_qual[i]:
+                    r1_seq[i] = b2
+                    r1_qual[i] = r2_qual[j]
+                    r1_changed += 1
+                elif self.mismatch_action == 'liberal':
+                    quals_equal.append((i, j, b1, b2))
+        
+        if quals_equal:
+            mean_qual1 = mean([ord(b) for b in r1_qual[r1_start:r1_end]])
+            mean_qual2 = mean([ord(b) for b in r2_qual[r2_start:r2_end]])
+            # Only make the corrections if one read is significantly better
+            # than the other.
+            # TODO: this method of determining whether one read is better
+            # than the other is crude - come up with something better.
+            diff = mean_qual1 - mean_qual2
+            if diff > 1:
+                # read1 is better than read2
+                for i, j, b1, b2 in quals_equal:
+                    r2_seq[j] = complement[b1]
+                    r2_qual[j] = r1_qual[i]
+                    r2_changed += 1
+            elif diff < -1:
+                # read2 is better than read1
+                for i, j, b1, b2 in quals_equal:
+                    r1_seq[i] = b2
+                    r1_qual[i] = r2_qual[j]
+                    r1_changed += 1
         
         if r1_changed or r2_changed:
             self.corrected_pairs += 1
@@ -447,6 +469,8 @@ class NextseqQualityTrimmer(Trimmer):
         self.base = base
     
     def __call__(self, read):
+        if len(read) == 0:
+            return read
         stop = nextseq_trim_index(read, self.cutoff, self.base)
         return self.subseq(read, end=stop)
 
@@ -460,6 +484,8 @@ class QualityTrimmer(Trimmer):
         self.base = base
 
     def __call__(self, read):
+        if len(read) == 0:
+            return read
         start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
         return self.subseq(read, start, stop)
 
@@ -474,6 +500,8 @@ class NEndTrimmer(Trimmer):
         self.end_trim = re.compile(r'N+$')
 
     def __call__(self, read):
+        if len(read) == 0:
+            return read
         sequence = read.sequence
         start_cut = self.start_trim.match(sequence)
         end_cut = self.end_trim.search(sequence)
@@ -512,6 +540,8 @@ class NonDirectionalBisulfiteTrimmer(object):
             self._rrbs_cutter = RRBSTrimmer(trim_3p)
     
     def __call__(self, read):
+        if len(read) == 0:
+            return read
         cutter = None
         if self._regex.match(read.sequence):
             cutter = self._non_directional_cutter
@@ -560,12 +590,15 @@ class MergeOverlapping(ReadPairModifier):
     
     def __call__(self, read1, read2):
         if read1.match and not read1.match.front and read2.match and not read2.match.front:
-            min_overlap = self.min_overlap
             l1 = len(read1.sequence)
             l2 = len(read2.sequence)
+            min_overlap = self.min_overlap
             if min_overlap <= 1:
                 min_overlap = max(2, round(self.min_overlap * min(l1, l2)))
-                
+            
+            if l1 < min_overlap or l2 < min_overlap:
+                return (read1, read2)
+            
             aligner = Aligner(read1.sequence, self.error_rate, SEMIGLOBAL)
             read2_rc = reverse_complement(read2.sequence)
             # TODO: Resolve overlap. We offer the user the choice to set mismatching bases to
@@ -604,7 +637,7 @@ class MergeOverlapping(ReadPairModifier):
 class Modifiers(object):
     def __init__(self, paired):
         self.modifiers = []
-        self.modifier_indexes = defaultdict(lambda: [])
+        self.modifier_indexes = {}
         self.paired = paired
     
     def add_modifier(self, mod_class, read=1|2, **kwargs):
@@ -634,18 +667,40 @@ class Modifiers(object):
     def _add_modifiers(self, mod_class, mods):
         i = len(self.modifiers)
         self.modifiers.append(mods)
-        self.modifier_indexes[mod_class].append(i)
+        if mod_class in self.modifier_indexes:
+            self.modifier_indexes[mod_class].append(i)
+        else:
+            self.modifier_indexes[mod_class] = [i]
         return i
     
     def get_modifiers(self, mod_class=None, read=None):
         if mod_class is None:
             mods = copy.copy(self.modifiers)
-        else:
+        elif mod_class in self.modifier_indexes:
             mods = [self.modifiers[i] for i in self.modifier_indexes[mod_class]]
-        if read is not None:
+        else:
+            mods = []
+        if mods and read:
             mods = [m[read-1] for m in mods if m[read-1] is not None]
         return mods
     
+    def has_modifier(self, mod_class):
+        return mod_class in self.modifier_indexes
+    
+    def get_adapters(self):
+        adapters = [None, None]
+        if self.has_modifier(AdapterCutter):
+            m1, m2 = self.get_modifiers(AdapterCutter)[0]
+            if m1:
+                adapters[0] = m1.adapters
+            if m2:
+                adapters[1] = m2.adapters
+        elif self.has_modifier(InsertAdapterCutter):
+            m = self.get_modifiers(InsertAdapterCutter)[0]
+            adapters[0] = [m.adapter1]
+            adapters[1] = [m.adapter2]
+        return adapters
+    
     def get_trimmer_classes(self):
         return [
             klass for klass in self.modifier_indexes.keys()
