@@ -40,13 +40,14 @@ from atropos import *
 check_importability()
 
 from atropos import __version__
-from atropos.seqio import (open_reader, UnknownFileType, Formatters, RestFormatter,
-                            InfoFormatter, WildcardFormatter, FormatError, Writers)
+from atropos.seqio import (
+    open_reader, UnknownFileType, Formatters, RestFormatter,
+    InfoFormatter, WildcardFormatter, Writers, BatchIterator)
 from atropos.adapters import AdapterParser, BACK
 from atropos.modifiers import *
 from atropos.filters import *
-from atropos.report import *
-from atropos.util import int_or_str
+from atropos.report import print_report
+from atropos.util import int_or_str, RandomMatchProbability
 
 import argparse
 from collections import defaultdict
@@ -96,13 +97,15 @@ def main(cmdlineargs=None, default_outfile="-"):
 def run_atropos(options, parser, default_outfile="-"):
     paired = validate_options(options, parser)
     reader, qualities, has_qual_file = create_reader(options, paired, parser)
-    modifiers, num_adapters = create_modifiers(options, paired, qualities, has_qual_file, parser)
+    modifiers, adapters1, adapters2 = create_modifiers(
+        options, paired, qualities, has_qual_file, parser)
     min_affected = 2 if options.pair_filter == 'both' else 1
     filters = create_filters(options, paired, min_affected)
     formatters, force_create = create_formatters(options, qualities, default_outfile)
     writers = Writers(force_create)
     
     logger = logging.getLogger()
+    num_adapters = len(adapters1) + len(adapters2)
     logger.info("Trimming %s adapter%s with at most %.1f%% errors in %s mode ...",
         num_adapters, 's' if num_adapters > 1 else '', options.error_rate * 100,
         { False: 'single-end', 'first': 'paired-end legacy', 'both': 'paired-end' }[paired])
@@ -118,7 +121,9 @@ def run_atropos(options, parser, default_outfile="-"):
     
     if options.threads is None:
         # Run single-threaded version
-        rc, summary = run_serial(reader, modifiers, filters, formatters, writers)
+        import atropos.serial
+        rc, summary = atropos.serial.run_serial(
+            reader, modifiers, filters, formatters, writers)
     else:
         # Run multiprocessing version
         import atropos.multicore
@@ -227,36 +232,59 @@ def get_argument_parser():
              "containing 3' adapters. New algorithms are being implemented and the default "
              "is likely to change. (adapter)")
     
+    # TODO: all the different matching options are pretty confusing. Either explain their
+    # usage better in the docs or find a way to simplify the choices.
+    
     # Arguments for adapter match
     group.add_argument("-e", "--error-rate", type=float, default=None,
         help="Maximum allowed error rate for adapter match (no. of errors divided by the length "
             "of the matching region). (0.1)")
+    group.add_argument("--indel-cost", type=int, metavar="COST", default=None,
+        help="Integer cost of insertions and deletions during adapter match. Substitutions always "
+             "have a cost of 1. (1)")
     group.add_argument("--no-indels", action='store_false', dest='indels', default=True,
         help="Allow only mismatches in alignments. "
             "(allow both mismatches and indels)")
     group.add_argument("-n", "--times", type=int, metavar="COUNT", default=1,
         help="Remove up to COUNT adapters from each read. (1)")
-    group.add_argument("-O", "--overlap", type=int, metavar="MINLENGTH", default=3,
-        help="If the overlap between the read and the adapter is shorter than "
-            "MINLENGTH, the read is not modified. Reduces the no. of bases "
-            "trimmed due to random adapter matches. (3)")
     group.add_argument("--match-read-wildcards", action="store_true", default=False,
         help="Interpret IUPAC wildcards in reads. (no)")
     group.add_argument("-N", "--no-match-adapter-wildcards", action="store_false",
         default=True, dest='match_adapter_wildcards',
         help="Do not interpret IUPAC wildcards in adapters. (no)")
+    # These two options are mutually exclusive. Currently, we give preference to -O to maintain
+    # compatibility with Cutadapt.
+    # Apparently, using a mutex sub-group causes a segfault under 3.5.2, at least when built
+    # by Traivs on linux. Commenting it out for now.
+    #adapter_mismatches = group.add_mutually_exclusive_group()
+    group.add_argument("-O", "--overlap", type=int, metavar="MINLENGTH", default=None,
+        help="If the overlap between the read and the adapter is shorter than "
+            "MINLENGTH, the read is not modified. Reduces the no. of bases "
+            "trimmed due to random adapter matches. (3)")
+    group.add_argument("--adapter-max-rmp", type=float, metavar="PROB", default=None,
+        help="If no minimum overlap (-O) is specified, then adapters are only matched "
+             "when the probabilty of observing k out of n matching bases is <= PROB. (1E-6)")
     
     # Arguments for insert match
+    group.add_argument("--adapter-pair", default=None, metavar="NAME1,NAME2",
+        help="When adapters are specified in files, this option selects a single pair to use "
+             "for insert-based adapter matching.")
+    group.add_argument("--insert-max-rmp", type=float, metavar="PROB", default=1E-6,
+        help="Overlapping inserts only match when the probablity of observing k of n "
+             "matching bases is <= PROB. (1E-6)")
     group.add_argument("--insert-match-error-rate", type=float, default=None,
         help="Maximum allowed error rate for insert match (no. of errors divided by the length "
             "of the matching region). (0.2)")
     group.add_argument("--insert-match-adapter-error-rate", type=float, default=None,
         help="Maximum allowed error rate for matching adapters after successful insert match "
              "(no. of errors divided by the length of the matching region). (0.2)")
-    group.add_argument("--correct-mismatches", choices=("best", "N"), default=None,
-        help="How to handle mismatches while aligning/merging; best: select the base with the best "
-             "quality; N: set the base to N. Note that if exactly one base is ambiguous, the "
-             "non-ambiguous base is always used.")
+    group.add_argument("--correct-mismatches", choices=("liberal", "conservative", "N"), default=None,
+        help="How to handle mismatches while aligning/merging. 'Liberal' and 'conservative' error "
+             "correction both involve setting the base to the one with the best quality. They differ "
+             "only when the qualities are equal -- liberal means set it to the base from the read with "
+             "the overall best median base quality, while conservative means to leave it unchanged. "
+             "'N' means to set the base to N. If exactly one base is ambiguous, the non-ambiguous base "
+             "is always used. (no error correction)")
     
     # Merging arguments
     group.add_argument("-R", "--merge-overlapping", action="store_true", default=False,
@@ -474,9 +502,7 @@ def get_argument_parser():
     
     # add subcommands
     parser.set_defaults(command=None)
-    subparsers = parser.add_subparsers(
-        dest="commands",
-        help='sub-command help')
+    subparsers = parser.add_subparsers(dest="commands", help='sub-command help')
     
     detect_parser = subparsers.add_parser("detect", help="Detect adapter sequences",
         usage="\natropos -se input.fastq detect\natropos -pe1 in1.fq -pe2 in2.fq detect",
@@ -486,8 +512,9 @@ def get_argument_parser():
         help="Size of k-mer used to scan reads for adapter sequences. (12)")
     detect_parser.add_argument("--max-adapters", type=int, default=None,
         help="The maximum number of candidate adapters to report. (none)")
-    detect_parser.add_argument("--include-unknown", action="store_true", default=False,
-        help="Also report sequences that are not in the list of known adapters.")
+    detect_parser.add_argument("--include-contaminants", choices=('all','known','unknown'), default='all',
+        help="What conaminants to search for: all, only known adapters/contaminants ('known'), "
+             "or only unknown contaminants ('unknown').")
     detect_parser.add_argument("--adapter-report", default="-",
         help="File in which to write the summary of detected adapters. (stdout)")
     detect_parser.add_argument("--known-contaminant", action="append", default=None,
@@ -496,6 +523,8 @@ def get_argument_parser():
     detect_parser.add_argument("--known-contaminants-file", default=None,
         help="File with known contaminants, one per line, with name and sequence "
              "separated by one or more tabs.")
+    detect_parser.add_argument("--no-cache-contaminants", action="store_true", default=False,
+        help="Don't cache contaminant list as '.contaminants' in the working directory.")
     
     error_parser = subparsers.add_parser("error", help="Estimate empirical error rate",
         usage="\natropos -se input.fastq error\natropos -pe in1.fq -pe2 in2.fq error",
@@ -522,6 +551,9 @@ def setup_logging(stdout=False, level="INFO"):
     logging.getLogger().addHandler(stream_handler)
 
 def validate_options(options, parser):
+    # TODO: most of these validation checks can be specified using
+    # functions passed to the 'type' parameter of add_argument
+    
     if sum(int(opt is not False) for opt in (options.mirna, options.bisulfite)) > 1:
         parser.error("Cannot specify more than one method-specific option")
     
@@ -573,17 +605,43 @@ def validate_options(options, parser):
             # Full paired-end trimming when both -p and -A/-G/-B/-U given
             # Read modifications (such as quality trimming) are applied also to second read.
             paired = 'both'
-
-    if options.aligner != 'adapter':
-        if options.aligner == 'insert':
-            if paired != 'both':
-                parser.error("Insert aligner only works with paired-end reads")
-                # TODO: should also be checking that there is exactly one 3' adapter for each read
-                # TODO: have the aligner tell us whether it can be used based on options?
-            if options.insert_match_error_rate is None:
-                options.insert_match_error_rate = options.error_rate or 0.2
-            if options.insert_match_adapter_error_rate is None:
-                options.insert_match_adapter_error_rate = options.error_rate or 0.2
+    
+    # If the user specifies a max rmp, that is used for determining the
+    # minimum overlap and -O is set to 1, otherwise -O is set to the old
+    # default of 3.
+    # TODO: This is pretty confusing logic - need to simplify
+    if options.overlap is not None and options.overlap < 1:
+        parser.error("The overlap must be at least 1.")
+    if options.indels and options.indel_cost is not None and options.indel_cost < 0:
+        parser.error("--indel-cost must be >= 0")
+    if options.adapter_max_rmp is not None and (options.adapter_max_rmp < 0 or options.adapter_max_rmp > 1.0):
+        parser.error("--adapter-max-rmp must be between [0,1].")
+    
+    if options.aligner == 'adapter':
+        if options.indels and options.indel_cost is None:
+            options.indel_cost = 1
+        if options.overlap is None:
+            if options.adapter_max_rmp is None:
+                options.overlap = 3
+            else:
+                options.overlap = 1
+    elif options.aligner == 'insert':
+        if paired != 'both':
+            parser.error("Insert aligner only works with paired-end reads")
+            # TODO: should also be checking that there is exactly one 3' adapter for each read
+            # TODO: have the aligner tell us whether it can be used based on options?
+        if options.insert_max_rmp < 0 or options.insert_max_rmp > 1.0:
+            parser.error("--insert-max-rmp must be between [0,1].")
+        if options.indels and options.indel_cost is None:
+            options.indel_cost = 3
+        if options.overlap is None:
+            options.overlap = 1
+            if options.adapter_max_rmp is None:
+                options.adapter_max_rmp = 1E-6
+        if options.insert_match_error_rate is None:
+            options.insert_match_error_rate = options.error_rate or 0.2
+        if options.insert_match_adapter_error_rate is None:
+            options.insert_match_adapter_error_rate = options.insert_match_error_rate
     
     if options.merge_overlapping and (
             paired != "both" or
@@ -697,12 +755,9 @@ def validate_options(options, parser):
     
     if options.error_rate is None:
         options.error_rate = 0.1
-    elif not (0 <= options.error_rate <= 1.):
+    elif not (0 <= options.error_rate <= 1.0):
         parser.error("The maximum error rate must be between 0 and 1.")
     
-    if options.overlap < 1:
-        parser.error("The overlap must be at least 1.")
-
     if options.cut:
         if len(options.cut) > 2:
             parser.error("You cannot remove bases from more than two ends.")
@@ -809,67 +864,23 @@ def create_reader(options, paired, parser, counter_magnitude="M"):
     
     return (reader, qualities, qualfile is not None)
 
-class BatchIterator(object):
-    def __init__(self, reader, size, max_reads=None):
-        self.reader = reader
-        self.iterable = enumerate(reader, 1)
-        self.size = size
-        self.max_reads = max_reads
-        self.done = False
-        self._empty_batch = [None] * size
-    
-    def __iter__(self):
-        return self
-    
-    def __next__(self):
-        if self.done:
-            raise StopIteration()
-        
-        try:
-            read_index, record = next(self.iterable)
-        except StopIteration:
-            self.close()
-            raise
-        
-        batch = copy.copy(self._empty_batch)
-        batch[0] = record
-        batch_index = 1
-        max_size = self.size
-        if self.max_reads:
-            max_size = min(max_size, self.max_reads - read_index + 1)
-        
-        while batch_index < max_size:
-            try:
-                read_index, record = next(self.iterable)
-                batch[batch_index] = record
-                batch_index += 1
-            except StopIteration:
-                self.close()
-                break
-        
-        if self.max_reads and read_index >= self.max_reads:
-            self.close()
-        
-        if batch_index == self.size:
-            return (batch_index, batch)
-        else:
-            return (batch_index, batch[0:batch_index])
-    
-    # py2x alias
-    next = __next__
-    
-    def close(self):
-        self.done = True
-        self.reader.close()
-
 def create_modifiers(options, paired, qualities, has_qual_file, parser):
-    adapter_parser = AdapterParser(
+    if options.adapter_max_rmp or options.aligner == 'insert':
+        match_probability = RandomMatchProbability()
+    
+    # Create adapters
+    parser_args = dict(
         colorspace=options.colorspace,
         max_error_rate=options.error_rate,
         min_overlap=options.overlap,
         read_wildcards=options.match_read_wildcards,
         adapter_wildcards=options.match_adapter_wildcards,
-        indels=options.indels)
+        indels=options.indels, indel_cost=options.indel_cost
+    )
+    if options.adapter_max_rmp:
+        parser_args['match_probability'] = match_probability
+        parser_args['max_rmp'] = options.adapter_max_rmp
+    adapter_parser = AdapterParser(**parser_args)
 
     try:
         adapters1 = adapter_parser.parse_multi(options.adapters, options.anywhere, options.front)
@@ -887,37 +898,46 @@ def create_modifiers(options, paired, qualities, has_qual_file, parser):
             options.cut_min == [] and options.cut_min2 == [] and \
             (options.minimum_length is None or options.minimum_length <= 0) and \
             options.maximum_length == sys.maxsize and \
-            not has_qual_file and \
-            options.max_n == -1 and not options.trim_n:
+            not has_qual_file and options.max_n == -1 and not options.trim_n:
         parser.error("You need to provide at least one adapter sequence.")
     
-    if options.aligner == 'insert' and (
-            not adapters1 or len(adapters1) > 1 or adapters1[0].where != BACK or
-            not adapters2 or len(adapters2) > 1 or adapters2[0].where != BACK):
-        parser.error("Insert aligner requires a single 3' adapter for each read")
+    if options.aligner == 'insert':
+        if options.adapter_pair and adapters1 and adapters2:
+            name1, name2 = options.adapter_pair.split(",")
+            adapters1 = [a for a in adapters1 if a.name == name1]
+            adapters2 = [a for a in adapters2 if a.name == name2]
+        if not adapters1 or len(adapters1) != 1 or adapters1[0].where != BACK or \
+                not adapters2 or len(adapters2) != 1 or adapters2[0].where != BACK:
+            parser.error("Insert aligner requires a single 3' adapter for each read")
     
     if options.debug:
         for adapter in adapters1 + adapters2:
             adapter.enable_debug()
     
+    # Create modifiers
     modifiers = Modifiers(paired)
-    
+            
     for op in options.op_order:
         if op == 'A' and (adapters1 or adapters2):
             # TODO: generalize this using some kind of factory class
             if options.aligner == 'insert':
-                aligner_args = dict(
-                    max_insert_mismatch_frac=options.insert_match_error_rate,
-                    min_adapter_match_frac=1.0 - options.insert_match_adapter_error_rate
-                )
+                # Use different base probabilities if we're trimming bisulfite data.
+                # TODO: this doesn't seem to help things, so commenting it out for now
+                #base_probs = dict(p1=0.33, p2=0.67) if options.bisulfite else dict(p1=0.25, p2=0.75)
                 modifiers.add_modifier(InsertAdapterCutter,
                     adapter1=adapters1[0], adapter2=adapters2[0], action=options.action,
-                    mismatch_action=options.correct_mismatches, **aligner_args)
+                    mismatch_action=options.correct_mismatches,
+                    max_insert_mismatch_frac=options.insert_match_error_rate,
+                    max_adapter_mismatch_frac=options.insert_match_adapter_error_rate,
+                    match_probability=match_probability,
+                    insert_max_rmp=options.insert_max_rmp)
             else:
-                modifiers.add_modifier_pair(AdapterCutter,
-                    dict(adapters=adapters1, times=options.times, action=options.action),
-                    dict(adapters=adapters2, times=options.times, action=options.action)
-                )
+                a1_args = a2_args = None
+                if adapters1:
+                    a1_args = dict(adapters=adapters1, times=options.times, action=options.action)
+                if adapters2:
+                    a2_args = dict(adapters=adapters2, times=options.times, action=options.action)
+                modifiers.add_modifier_pair(AdapterCutter, a1_args, a2_args)
         elif op == 'C' and (options.cut or options.cut2):
             modifiers.add_modifier_pair(UnconditionalCutter,
                 dict(lengths=options.cut),
@@ -940,7 +960,9 @@ def create_modifiers(options, paired, qualities, has_qual_file, parser):
             elif options.bisulfite == "rrbs":
                 modifiers.add_modifier(RRBSTrimmer)
             elif options.bisulfite in ("epignome", "truseq"):
-                modifiers.add_modifier(TruSeqBisulfiteTrimmer)
+                # Trimming leads to worse results
+                #modifiers.add_modifier(TruSeqBisulfiteTrimmer)
+                pass
             elif options.bisulfite == "swift":
                 modifiers.add_modifier(SwiftBisulfiteTrimmer)
         else:
@@ -981,7 +1003,7 @@ def create_modifiers(options, paired, qualities, has_qual_file, parser):
             min_overlap=options.merge_min_overlap,
             error_rate=options.error_rate)
     
-    return (modifiers, len(adapters1) + len(adapters2))
+    return (modifiers, adapters1, adapters2)
 
 def create_filters(options, paired, min_affected):
     filters = Filters(FilterFactory(paired, min_affected))
@@ -1064,44 +1086,5 @@ def create_formatters(options, qualities, default_outfile):
     
     return (formatters, force_create)
 
-def run_serial(reader, modifiers, filters, formatters, writers):
-    try:
-        n = 0
-        total_bp1 = 0
-        total_bp2 = 0
-        
-        for batch_size, batch in reader:
-            n += batch_size
-            result = defaultdict(lambda: [])
-            for record in batch:
-                reads, bp = modifiers.modify(record)
-                total_bp1 += bp[0]
-                total_bp2 += bp[1]
-                dest = filters.filter(*reads)
-                formatters.format(result, dest, *reads)
-            result = dict((path, "".join(strings))
-                for path, strings in result.items())
-            writers.write_result(result)
-        
-        return (0, Summary(
-            collect_process_statistics(n, total_bp1, total_bp2, modifiers, filters, formatters),
-            summarize_adapters(modifiers.get_modifiers(AdapterCutter)),
-            modifiers.get_trimmer_classes()
-        ).finish())
-    
-    except KeyboardInterrupt as e:
-        logging.getLogger().error("Interrupted")
-        return (130, None)
-    except IOError as e:
-        if e.errno == errno.EPIPE:
-            return (1, None)
-        raise
-    except (FormatError, EOFError) as e:
-        logging.getLogger().error("Atropos error", exc_info=True)
-        return(1, None)
-    finally:
-        reader.close()
-        writers.close()
-
 if __name__ == '__main__':
     main()
